@@ -6,7 +6,7 @@
 /* Allow compilation on Windows (thanks, Fabrice Popineau). */
 #ifdef WIN32
 # ifdef __MINGW32__
-#  include <windows.h>
+#  include <winsock2.h>
 # else
 #  include <win32lib.h>
 # endif
@@ -38,6 +38,11 @@ double good_strtod(const char *s, char **endptr);
 }
 /* Get rid of a possible inline macro under C++. */
 # define inline inline
+
+/* Ignore `noexcept` if compiler is too old. */
+#if __cplusplus < 201103L
+#define noexcept
+#endif
 #endif
 
 #endif /* LCDF_TYPETOOLS_CONFIG_H */
